# coding=utf8
import collections

__author__ = 'Alexander.Li'


# Modular arithmetic ##########################################################

# Alexander.Li: Next functions from https://github.com/andreacorbellini/ecc

EllipticCurve = collections.namedtuple('EllipticCurve', 'name p a b g n h')

curve = EllipticCurve(
    'secp256k1',
    # Field characteristic.
    p=0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f,
    # Curve coefficients.
    a=0,
    b=7,
    # Base point.
    g=(0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798,
       0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8),
    # Subgroup order.
    n=0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141,
    # Subgroup cofactor.
    h=1,
)


# Functions that work on curve points #########################################

def inverse_mod(k, p):
    """Returns the inverse of k modulo p.
    This function returns the only integer x such that (x * k) % p == 1.
    k must be non-zero and p must be a prime.
    """
    if k == 0:
        raise ZeroDivisionError('division by zero')

    if k < 0:
        # k ** -1 = p - (-k) ** -1  (mod p)
        return p - inverse_mod(-k, p)

    # Extended Euclidean algorithm.
    s, old_s = 0, 1
    t, old_t = 1, 0
    r, old_r = p, k

    while r != 0:
        quotient = old_r // r
        old_r, r = r, old_r - quotient * r
        old_s, s = s, old_s - quotient * s
        old_t, t = t, old_t - quotient * t

    gcd, x, y = old_r, old_s, old_t

    assert gcd == 1
    assert (k * x) % p == 1

    return x % p


def is_on_curve(point):
    """Returns True if the given point lies on the elliptic curve."""
    if point is None:
        # None represents the point at infinity.
        return True

    x, y = point

    return (y * y - x * x * x - curve.a * x - curve.b) % curve.p == 0


def point_neg(point):
    """Returns -point."""
    assert is_on_curve(point)

    if point is None:
        # -0 = 0
        return None

    x, y = point
    result = (x, -y % curve.p)

    assert is_on_curve(result)

    return result


def point_add(point1, point2):
    """Returns the result of point1 + point2 according to the group law."""
    assert is_on_curve(point1)
    assert is_on_curve(point2)

    if point1 is None:
        # 0 + point2 = point2
        return point2
    if point2 is None:
        # point1 + 0 = point1
        return point1

    x1, y1 = point1
    x2, y2 = point2

    if x1 == x2 and y1 != y2:
        # point1 + (-point1) = 0
        return None

    if x1 == x2:
        # This is the case point1 == point2.
        m = (3 * x1 * x1 + curve.a) * inverse_mod(2 * y1, curve.p)
    else:
        # This is the case point1 != point2.
        m = (y1 - y2) * inverse_mod(x1 - x2, curve.p)

    x3 = m * m - x1 - x2
    y3 = y1 + m * (x3 - x1)
    result = (x3 % curve.p,
              -y3 % curve.p)

    assert is_on_curve(result)

    return result


def scalar_mult(k, point):
    """Returns k * point computed using the double and point_add algorithm."""
    assert is_on_curve(point)

    if k % curve.n == 0 or point is None:
        return None

    if k < 0:
        # k * point = -k * (-point)
        return scalar_mult(-k, point_neg(point))

    result = None
    addend = point

    while k:
        if k & 1:
            # Add.
            result = point_add(result, addend)

        # Double.
        addend = point_add(addend, addend)

        k >>= 1

    assert is_on_curve(result)

    return result


import hashlib
import random
from sys import version_info
from .curv import curve, scalar_mult, inverse_mod, point_add


def hash_message(message):
    """Returns the truncated SHA521 hash of the message."""
    message_hash = hashlib.sha512(message).digest()

    if version_info.major != 2:
        e = int.from_bytes(message_hash, 'big')
    else:
        e = int(message_hash.encode('hex'), 16)

    # FIPS 180 says that when a hash needs to be truncated, the rightmost bits
    # should be discarded.
    z = e >> (e.bit_length() - curve.n.bit_length())

    assert z.bit_length() <= curve.n.bit_length()

    return z


def sign_message(private_key, message):
    z = hash_message(message)

    r = 0
    s = 0

    while not r or not s:
        k = random.randrange(1, curve.n)
        x, y = scalar_mult(k, curve.g)

        r = x % curve.n
        s = ((z + r * private_key) * inverse_mod(k, curve.n)) % curve.n

    return (r, s)


def verify_signature(public_key, message, signature):
    z = hash_message(message)
    r, s = signature

    w = inverse_mod(s, curve.n)
    u1 = (z * w) % curve.n
    u2 = (r * w) % curve.n

    x, y = point_add(scalar_mult(u1, curve.g),
                     scalar_mult(u2, public_key))

    if (r % curve.n) == (x % curve.n):
        return True
    else:
        return False





from setuptools import setup, find_packages
from distutils.extension import Extension
from Cython.Build import cythonize, build_ext

version = '0.0.8'


setup(name='secp256k1py',
      version=version,
      description="Python version secp256k1 keypair generator signature and verify, ecdh secret sharing, for human mind",
      long_description="""\
Python version secp256k1 keypair generator signature and verify, ecdh secret sharing, for human mind""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='ecc,ecdh,secp256k1,signature,verification',
      author='Alexander.Li',
      author_email='superpowerlee@gmail.com',
      url='https://github.com/ipconfiger/secp256k1-py',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      package_data={
              'secp256k1py': ['*.pyx',]
      },
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          "point",
          "salsa20>=0.3.0"
      ],
      cmdclass={'build_ext': build_ext},
      ext_modules=cythonize([Extension('*', ['secp256k1py/curv.pyx'])]),
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



# coding=utf8

import base64
import hashlib
import random
from os import urandom
from sys import version_info
import secp256k1py.functions
from salsa20 import Salsa20_xor


class PrivateKey():
    def __init__(self, _d):
        self.d = _d

    @classmethod
    def restore(cls, hex_str):
        if version_info.major != 2:
            return cls(int(hex_str, 16))
        else:
            return cls(long(hex_str, 16))

    def __repr__(self):
        return hex(self.d)[2:-1]

    def generate_secret(self, publickey):
        """
        生成共享秘密
        :param publickey:
        :return:
        """
        point = secp256k1py.functions.scalar_mult(self.d, publickey.Q)
        x, y = point
        return "%s%s" % (hex(x)[2:-1], hex(y)[2:-1])

    def sign(self, message):
        """
        签名消息
        :param message:
        :return:
        """
        point = secp256k1py.functions.sign_message(self.d, message)
        x, y = point
        return "%s%s" % (hex(x)[2:-1], hex(y)[2:-1])

    def decrypt(self, publicKey, b64encrypted, b64iv):
        """
        解压数据
        :param publicKey:
        :param b64encrypted:
        :param b64iv:
        :return:
        """
        uncompress_key = self.generate_secret(publicKey).decode('hex')
        key = hashlib.sha256(uncompress_key).digest()
        raw_enc_bytes = base64.urlsafe_b64decode(b64encrypted)
        iv = base64.urlsafe_b64decode(b64iv)
        return Salsa20_xor(raw_enc_bytes, iv, key)


class PublicKey():
    def __init__(self, _q):
        self.Q = _q

    @classmethod
    def restore(cls, hex_str):
        if version_info.major != 2:
            point = (
                int(hex_str[:64], 16),
                int(hex_str[64:], 16)
            )
        else:
            point = (
                long(hex_str[:64], 16),
                long(hex_str[64:], 16)
            )
        return cls(point)

    def verify(self, message, signature):
        """
        对消息验签
        :param message:
        :param signature:
        :return:
        """
        if version_info.major != 2:
            point = (
                int(signature[:64], 16),
                int(signature[64:], 16)
            )
        else:
            point = (
                long(signature[:64], 16),
                long(signature[64:], 16)
            )
        return secp256k1py.functions.verify_signature(self.Q, message, point)

    def encrypt(self, privateKey, message):
        """
        用共享秘密加密数据
        :param privateKey:
        :return:
        """
        uncompress_key = privateKey.generate_secret(self).decode('hex')
        key = hashlib.sha256(uncompress_key).digest()
        iv = urandom(8)
        enc = Salsa20_xor(message, iv, key)
        b64_enc = base64.urlsafe_b64encode(enc)
        b64_iv = base64.urlsafe_b64encode(iv)
        return dict(
            enc=b64_enc,
            iv=b64_iv
        )

    def __repr__(self):
        x, y = self.Q
        return "%s%s" % (hex(x)[2:-1], hex(y)[2:-1])


class KeyPair():
    def __init__(self, private, public):
        self.privateKey = private
        self.publicKey = public


def make_keypair():
    """Generates a random private-public key pair."""
    private_key = random.randrange(1, secp256k1py.functions.curve.n)
    public_key = secp256k1py.functions.scalar_mult(private_key, secp256k1py.functions.curve.g)
    return KeyPair(PrivateKey(private_key), PublicKey(public_key))
